# -*- perl -*-

# Copyright (c) 2002 by Jeff Weisberg
# Author: Jeff Weisberg <argus @ tcp4me.com>
# Date: 2002-Apr-03 18:25 (EST)
# Function: testing of UDP services
#
# $Id: UDP.pm,v 1.32 2002/12/01 21:30:56 jaw Exp $


package UDP;
@ISA = qw(Service);

use Fcntl;
use Socket;
use SNMP;
use DNS;
# use Domain;
use POSIX qw(:errno_h);
BEGIN {
    eval { require Socket6; import Socket6; };
}

$doc = {
    package => __PACKAGE__,
    file    => __FILE__,
    isa     => [qw(Service MonEl BaseIO)],
    methods => {},
    html   => 'services',
    fields => {
      udp::port => {
	  descr => 'UDP port to test',
	  attrs => ['config'],
      },
      udp::addr => {
	  descr => 'encoded IP address',
      },
      udp::send => {
	  descr => 'text to send once connected',
	  attrs => ['config'],
      },
      udp::rbuffer => {
	  descr => 'read buffer',
      },
      udp::wbuffer => {
	  descr => 'write buffer',
      },
	
    },
};

%config =
(
 # not supported by all radius servers
 RADIUS => {
     port => 1645, send => pack( "CCSx20", 12, $$, 20 ),
     timeout => 10,
 },
 
 DNS => { # RFC 1035, op=status, not supported by all DNS servers
     port => 53,   send => pack("n CC nnnn", $$, 0x10,0, 0,0,0,0),
     timeout => 12,
 },

 DNSQ => { # for DNS servers that don't respond to status queries
     port => 53, timeout => 12,
     send => pack( "n CC nnnn  Cnn", $$, 1,0, 1,0,0,0, 0,255,1),
     # op=query, recurse, 1 query, . ANY IN
     expect => '^...(\x80|\x00)',	# rcode = No error condition
     # I could test that it returned 13 root servers, but if I did that
     # the very next day John would announce he's adding an 'N'
 },
 
 # test server authority for zone
 Domain => {
     port => 53, timeout => 12,
 },
 
 SNMP => {
     port => 161,  timeout => 5, 
 },

 NTP => { # RFC 2030
     port => 123,  timeout => 10,
     send => pack("CCCC x64", 0x23,0,0,0),
 },

 # NFS = RFC 1094; RPC = RFC 1057; XDR = RFC 1014
 # NFSv2 - NFSPROC_NULL
 NFS => {
     port => 2049, timeout => 10,
     send => pack( "NN NN NN x16", $$, 0,  2, 100003, 2, 0),
     # xid, type, rpcver, prog, ver, func, cred(flavor, len, null), verf(flavor, len, null)
 },

 NFSv3 => {
     port => 2049, timeout => 10,
     send => pack( "NN NN NN x16", $$, 0,  2, 100003, 3, 0),
 },

 );

my $PROTO_UDP = getprotobyname('udp');

sub probe {
    my $name = shift;

    return ( $name =~ /^UDP/ ) ? [ 3, \&config ] : undef;
}

sub config {
    my $me = shift;
    my $cf = shift;
    my( $name, $base );

    bless $me if( ref($me) eq 'Service' );

    $name = $me->{name};
    $name =~ s/^UDP\/?//;
    $base = $name;
    $base =~ s/\/.*//;

    $me->{srvc}{timeout}     = $config{$base}{timeout};
    
    $me->init_from_config( $doc, 'udp' );

    $me->{udp}{port}        ||= $config{$base}{port};
    $me->{udp}{send}        ||= $config{$base}{send};
    $me->{test}{expect}       = $config{$base}{expect};
    
    $me->{udp}{hostname}    = $me->{srvc}{hostname};
    
    if( $name ){
	$me->{uname} = "${name}_$me->{srvc}{hostname}";
    }else{
	$me->{uname} = "UDP_$me->{udp}{port}_$me->{srvc}{hostname}";
    }

    $me->{udp}{addr} = ::resolve( $me->{srvc}{hostname} );
    unless( $me->{udp}{addr} ){
	return $cf->error( "cannot resolve host $me->{srvc}{hostname}" );
    }

    $me->{label_right_maybe} ||= $name;

    unless( $me->{udp}{port} ){
	return $cf->error( "Incomplete specification or unknown protocol for Service $name" );
    }

    $me;
}

sub start {
    my $me = shift;
    my( $fh, $i, $ip, $ipv );

    $me->{fd} = $fh = BaseIO::anon_fh();
    $ip = $me->{udp}{addr};
    $me->Service::start();
    
    if( length($ip) == 4 ){
	$i = socket($fh, PF_INET, SOCK_DGRAM, $PROTO_UDP);
    }else{
	$i = socket($fh, PF_INET6, SOCK_DGRAM, $PROTO_UDP);
	$ipv = ' IPv6';
    }
    unless($i){
	my $m = "socket failed: $!";
	::sysproblem( "UDP $m" );
	$me->debug( $m );
	return $me->done();
    }

    $me->baseio_init();
    
    $me->debug( "UDP$ipv start" );
    if( length($ip) == 4 ){
	$i = bind($fh, sockaddr_in(0, INADDR_ANY) );
    }else{
	# XXX - Socket6.pm says this is 'inaddr6_any', which is the typo?
	$i = bind($fh, pack_sockaddr_in6(0, in6addr_any) );
    }
    unless($i){
	my $m = "bind failed: $!";
	::sysproblem( "UDP $m" );
	$me->debug( $m );
	return $me->done();
    }

    # And they made the plate of the holy crown of pure gold, and wrote upon it a writing
    #   -- exodus 39:30
    if( length($ip) == 4 ){
	$i = send( $fh, $me->{udp}{send}, 0, sockaddr_in($me->{udp}{port}, $ip) );
    }else{
	$i = send( $fh, $me->{udp}{send}, 0, pack_sockaddr_in6($me->{udp}{port}, $ip) );
    }
    unless( $i ){
	# QQQ, should I try to sort out errors, down on some, sysproblem on others?
	$me->isdown( "UDP send failed: $!" );
	return;
    }

    $me->debug( "UDP sent data" );

    $me->wantread(1);
    $me->wantwrit(0);
    $me->settimeout( $me->{srvc}{timeout} );
    $me->{srvc}{state} = "waiting";

}

sub readable {
    my $me = shift;
    my( $fh, $i, $l, $e );

    $fh = $me->{fd};
    $i = recv($fh, $l, 8192, 0);
    return $me->isdown( "UDP recv failed: $!" )
	unless defined($i);

    $me->debug( "UDP recv data" );
    $me->{udp}{rbuffer} = $l;		# for debugging
    
    # yet I will read the writing unto the king, and make known to him the interpretation
    #   -- daniel 5:17

    return $me->generic_test( $l );
}

sub timeout {
    my $me = shift;

    $me->isdown( "UDP timeout" );
}

################################################################
# and also object methods
################################################################

# sub done {
#     my $me = shift;
# 
#     $me->Service::done();
# }
# 
# sub isdown {
#     my $me = shift;
#     my $reason = shift;
# 
#     $me->Service::isdown($reason);
# }
# 
# sub isup {
#     my $me = shift;
# 
#     $me->Service::isup();
# }

sub about_more {
    my $me = shift;
    my $ctl = shift;
    my( $k, $v );

    $me->Service::about_more($ctl);

    foreach $b ('udp'){
	next unless $me->{$b};
	foreach $k (sort keys %{$me->{$b}}){
	    $v = $me->{$b}{$k};
	    next unless defined $v;
	    if( ref($v) ){
		$v = "#<REF>";
	    }else{
		# QQQ - or some other encoding scheme?
		$v = ::xxx_encode( $v );
	    }
	    $ctl->write( "${b}::$k:\t$v\n" );
	}
    }
}

sub webpage_more {
    my $me = shift;
    my $fh = shift;
    my( $k, $v );
    
    foreach $k (qw(hostname port)){
	$v = $me->{udp}{$k};
	print $fh "<TR><TD>$k</TD><TD>$v</TD></TR>\n" if defined($v);
    }
}

################################################################
# global config
################################################################
Doc::register( $doc );
push @Service::probes, \&probe;

1;

